/**
  Copyright (C) 1998 XYZ, Inc. All rights Reserved.
  XXX   %Z%%M% version %I% date:%G%
  Component:  LocalSPKIRepository
  Package: repository
  File:    LocalSPKIRepository.java
*/

package fi.hut.tcm.repository;
import fi.hut.tcm.spki.*;
import fi.hut.tcm.spki.impl.*;
import java.io.*;
import java.security.cert.Certificate;
import java.util.Vector;
import java.security.*;
import sun.security.util.PropertyExpander;

/**
  LocalSPKIRepository
  
  The LocalSPKIRepository uses a file to store the certificates.
*/
class LocalSPKIRepository
  extends SPKICertificateRepository {

  /**
    Indicates if the contents of this repository have been modified.
  */
  private boolean modified = false;
  
  /**
    Indexes the certificates by issuer.
  */
  private CertTable byIssuer;
  
  /**
    Indexes the certificates by subject.
  */
  private CertTable bySubject;
  
  /**
    The name of the file where the certificates are stored on the disk.
  */
  private String fileName;
  
  /**
    Creates a new LocalSPKIRepository.
  */
  public LocalSPKIRepository() {
  
  //+ LocalSPKIRepository::LocalSPKIRepository( )
    // FIX: is this what we want to do?
    byIssuer = new CertTable();
    bySubject = new CertTable();

    fileName = (String) AccessController.doPrivileged(new PrivilegedAction() {
      public Object run() {
	return Security.getProperty("certrepository.file");
      };
    });
    if (fileName == null) {
      try {
        fileName = PropertyExpander.expand("${user.home}/CertRepository.dat");
      } catch (Exception e) {
	// FIX: do something!
	System.err.println("LocalCertificateRepository: unable to create repository file name." + e.toString());
      }
    }
  //- LocalSPKIRepository::LocalSPKIRepository( )
  }
  
  /**
    Initializes the repository.
  */
  public void init() {
  
  //+ LocalSPKIRepository::init( )
    try {
      load();
    } catch (Exception e) {
      // FIX: do something!
      System.err.println("LocalCertificateRepository: load failed." + e.toString());
    }
    
  //- LocalSPKIRepository::init( )
  }
  
  /**
    Adds the cert to the repository.
  */
  public void add(Certificate cert) 
    throws SPKIException  {
  
  //+ LocalSPKIRepository::add(Certificate)
    SPKICertificate scert;
      
    if (cert instanceof SPKICertificate)
      scert = (SPKICertificate) cert;
    else
      throw new SPKIException("Certificate not a SPKI certificate: " + cert.toString());
    try {
      Issuer tmp = new Issuer(scert.getIssuer());
      byIssuer.put(String.valueOf(tmp.toString().hashCode()), scert);
      bySubject.put(String.valueOf(scert.getSubject().toString().hashCode()), scert);
      modified = true;
    } catch (Exception e) {
      throw new SPKIException("Certificate does not have an issuer or a subject." + e.toString());
    }
  //- LocalSPKIRepository::add(Certificate)
  }
  
  /**
    Removes the cert from the repository.
  */
  public void remove(Certificate cert)
    throws Exception  {
  
  //+ LocalSPKIRepository::remove(Certificate)
    SPKICertificate scert;
      
    if (cert instanceof SPKICertificate)
      scert = (SPKICertificate) cert;
    else
      throw new SPKIException("Certificate not a SPKI certificate: " + cert.toString());
    byIssuer.remove(String.valueOf(scert.getIssuer().toString().hashCode()), scert);
    bySubject.remove(String.valueOf(scert.getSubject().toString().hashCode()), scert);
    modified = true;
  //- LocalSPKIRepository::remove(Certificate)
  }
  
  /**
    Returns all certificates that have been issued by the issuer.
  */
  public Vector getCertsByIssuer(Issuer issuer) {
  
  //+ LocalSPKIRepository::getCertsByIssuer(Issuer)
    return byIssuer.get(String.valueOf(issuer.toString().hashCode()));
  //- LocalSPKIRepository::getCertsByIssuer(Issuer)
  }
  
  /**
    Returns all certificates that have been issued to the subject.
  */
  public Vector getCertsBySubject(Subject subject) {
  
  //+ LocalSPKIRepository::getCertsBySubject(Subject)
    return bySubject.get(String.valueOf(subject.toString().hashCode()));
  //- LocalSPKIRepository::getCertsBySubject(Subject)
  }
  
  /**
    Saves the repository.
  */
  public void finalize() {
  
  //+ LocalSPKIRepository::finalize( )
    if (modified) {
      try {
	save();
      } catch (Exception e) {
	// FIX: do something!
	System.err.println("LocalCertificateRepository: save failed." + e.toString());
      }
    }
  //- LocalSPKIRepository::finalize( )
  }
  
  /**
    Loads the CertTables from a file.
  */
  public void load()
    throws OptionalDataException, ClassNotFoundException, 
      StreamCorruptedException, FileNotFoundException, IOException  {
  
  //+ LocalSPKIRepository::load( )
    // FIX: the load does not work
    try {
      SPKICertificateFactory factory = new SpkiCertFactory();
      KeyStore ks = KeyStore.getInstance("JKS");
      ks.load(new FileInputStream(new String(PropertyExpander.expand("${user.home}/.keystore"))), "passwd".toCharArray());
      
      java.security.cert.Certificate certif = ks.getCertificate("Self");
      java.security.KeyPair kp = new java.security.KeyPair(certif.getPublicKey(), 
							   (PrivateKey)ks.getKey("Self", "passwd".toCharArray()));
      fi.hut.tcm.spki.Tag tag = new fi.hut.tcm.spki.impl.Tag();
      tag.add(new fi.hut.tcm.spki.impl.TagStar());
      add(factory.makeCert(kp, ks.getCertificate("Admin").getPublicKey(), true, tag));
    } catch (Exception e) {
      throw new IOException(e.toString());
    }
    /*
    FileInputStream fsin = new FileInputStream(fileName);
    ObjectInputStream objin = new ObjectInputStream(fsin);
    byIssuer = (CertTable) objin.readObject();
    bySubject = (CertTable) objin.readObject();
    fsin.close();
    */
  //- LocalSPKIRepository::load( )
  }
  
  /**
    Stores the CertTables to a file.
  */
  public void save()
    throws IOException  {
  
  //+ LocalSPKIRepository::save( )
    FileOutputStream fsout = new FileOutputStream(fileName);
    ObjectOutputStream objout = new ObjectOutputStream(fsout);
    objout.writeObject(byIssuer);
    objout.writeObject(bySubject);
    fsout.close();
  //- LocalSPKIRepository::save( )
  }
  
} // LocalSPKIRepository

// +*************** UNUSED CODE ***************
// -*************** UNUSED CODE ***************
