/**
  Copyright (C) 1998 XYZ, Inc. All rights Reserved.
  XXX   %Z%%M% version %I% date:%G%
  Component:  CertData
  Package: format
  File:    CertData.java
*/

package fi.hut.tcm.spki.format;
import java.util.*;
import java.net.URL;
import java.security.PublicKey;
import fi.hut.tcm.spki.SPKIException;
import fi.hut.tcm.spki.ByteString;
import fi.hut.tcm.spki.Component;
import fi.hut.tcm.spki.Validity;
import fi.hut.tcm.spki.format.*;
import fi.hut.tcm.spki.impl.*;

/**
  CertData
  
  Contains different certificate formats. May hold sequences of 
  certificates as well as single ones. When several formats coexist, 
  changing one means that others should be dropped or recreated.
*/
public class CertData
  extends Object {

  /**
  */
  private Vector formats;
  
  /**
    Always at least one format.
  */
  public CertData(Format first) {
  
  //+ CertData::CertData(Format)
    super();
    formats = new Vector();
    formats.addElement(first);
  //- CertData::CertData(Format)
  }
  
  /**
    Adds new certificate formats. It is the responsibility to assure that 
    the formats are consistent with each other. If this kind of format 
    already exists, the new format replaces the old one.
  */
  public void addFormat(Format addThis) {
  
  //+ CertData::addFormat(Format)
    if (getFormat(addThis) != null)
      removeFormat(addThis);
    formats.addElement(addThis);
  //- CertData::addFormat(Format)
  }
  
  /**
    Returns the format that is of the same type than the argument. If not 
    found, null is returned. Returns GenericFormat if the argument is 
    either of it's possible subclasses.
  */
  public Format getFormat(Format type) {
  
  //+ CertData::getFormat(Format)
    for (int i = 0; i < formats.size(); i++)
      if (type instanceof GenericFormat) {
	if (formats.elementAt(i) instanceof GenericFormat)
	  return (Format) formats.elementAt(i);
      }
      else
	if (type.getClass().isInstance(formats.elementAt(i)))
	  return (Format) formats.elementAt(i);
    return null;
  //- CertData::getFormat(Format)
  }
  
  /**
    Removes the format that is of the same type than the argument. Throws 
    the exception if such format was not found.
  */
  public void removeFormat(Format type)
    throws IllegalArgumentException  {
  
  //+ CertData::removeFormat(Format)
      for (int i = 0; i < formats.size(); i++)
	if (type instanceof GenericFormat) {
	  if (formats.elementAt(i) instanceof GenericFormat) {
	    formats.removeElementAt(i);
	    return;
	  }
	}
	else
	  if (type.getClass().isInstance(formats.elementAt(i))) {
	    formats.removeElementAt(i);
	    return;
	  }
      throw new IllegalArgumentException();
  //- CertData::removeFormat(Format)
  }
  
  /**
  */
  public boolean isValid() {
  
  //+ CertData::isValid( )
    // XXX Fix this
    System.out.println("CertData: not implemented yet.");
    return false; 
  //- CertData::isValid( )
  }
  
  /**
    Parses Spki transfer representation. Throws NoSuchFieldException if 
    CanonicalSpki is not among the formats and IllegalArgumentException if 
    CanonicalSpki parsing fails.
  */
  public void parseCanonicalSpki()
    throws NoSuchFieldException,IllegalArgumentException  {
  
  //+ CertData::parseCanonicalSpki( )
      SpkiParserVisitor parser = new SpkiParserVisitor();
      CanonicalSpki bin = (CanonicalSpki) getFormat(new CanonicalSpki());
      
      if (bin == null)
	throw new NoSuchFieldException();
      
      addFormat(parser.run(bin)); // May throw IllegalArgumentException
  //- CertData::parseCanonicalSpki( )
  }
  
  /**
    Makes Spki transfer representation. Throws NoSuchFieldException if 
    GenericFormat does not exist and IllegalArgumentException if 
    GenericFormat was bad.
  */
  public void makeCanonicalSpki()
    throws NoSuchFieldException,IllegalArgumentException  {
  
  //+ CertData::makeCanonicalSpki( )
      SpkiMakerVisitor maker = new SpkiMakerVisitor();
      // We will instantiate Sequence, because GenericFormat is an interface.
      GenericFormat gen = (GenericFormat) getFormat(new Sequence());
      
      if (gen == null)
	throw new NoSuchFieldException();
      
      addFormat(maker.run((Composite) gen)); // May throw IllegalArgumentException
  //- CertData::makeCanonicalSpki( )
  }
  
  /**
    Tests if GenericFormat is single certificate instead of a sequence. 
    Throws the exception if GenericFormat does not exist.
  */
  public boolean isSingleCert()
    throws NoSuchMethodException  {
  
  //+ CertData::isSingleCert( )
      GenericFormat gen;
      
      if ((gen = (GenericFormat) getFormat(new Sequence())) == null)
	throw new NoSuchMethodException();
      return gen.isSingleCert();
  //- CertData::isSingleCert( )
  }
  
  /**
    Reduces a sequence to a single certificate. Throws 
    NoSuchMethodException if GenericFormat does not exist.
  */
  public void reduceSequence()
    throws NoSuchMethodException  {
  
  //+ CertData::reduceSequence( )
      if (!isSingleCert()) {
	Sequence seq = (Sequence) getFormat(new Sequence());
	HashMap table = new HashMap();
	Stack result = new Stack();

	while (seq.size() == 1) {
	  SeqEnt first = (SeqEnt) seq.get(0);
	  SeqEnt second = (SeqEnt) seq.get(1);
	  if (first instanceof Cert) {
	    if (second instanceof Op) {
	      // XXX if hash, save first and hash, remove second, else ??
	      continue;
	    }
	    if (second instanceof Sig) {
	      // XXX check validity of sig (is this the place to do it?)
	      continue;
	    }
	    // XXX reduction of first and pop(result)
	  }
	  if (first instanceof PubKey) {
	    // XXX
	  }
	  if (first instanceof Op) {
	    // XXX
	  }
	  if (first instanceof Sig) {
	    // XXX
	  }
	}
	// XXX yksi jonossa ja mahd. result
      }
  //- CertData::reduceSequence( )
  }
  
  /**
    Adds a new entity at the end of a sequence. Throws the exception if 
    GenericFormat is a single certificate.
  */
  public void addToSequence(SeqEnt entity)
    throws NoSuchMethodException  {
  
  //+ CertData::addToSequence(SeqEnt)
      if (isSingleCert())
	throw new NoSuchMethodException();
      else
	((Sequence) getFormat(new Sequence())).add((Composite) entity);
  //- CertData::addToSequence(SeqEnt)
  }
  
  /**
    Returns an iterator of the sequence entities. Throws the exception 
    if GenericFormat is a single certificate.
  */
  public Iterator iterator()
    throws NoSuchMethodException  {
  
  //+ CertData::iterator( )
      if (isSingleCert())
	throw new NoSuchMethodException();
      else
	return ((Sequence) getFormat(new Sequence())).iterator();
  //- CertData::iterator( )
  }
  
  /**
    Removes an entity from a sequence. Throws NoSuchMethodException if 
    GenericFormat is a single certificate and NoSuchFieldException if this 
    entity does not exist in a sequence.
  */
  public void removeSequenceEntity(SeqEnt removeThis)
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::removeSequenceEntity(SeqEnt)
      if (isSingleCert())
	throw new NoSuchMethodException();
      else
	if (((Sequence) getFormat(new Sequence())).
	    remove((Component) removeThis) == false)
	  throw new NoSuchFieldException();
  //- CertData::removeSequenceEntity(SeqEnt)
  }
  
  /**
    Returns the version number of the certificate type. At the moment this 
    is always 0.
  */
  public int getVersion() {
  
  //+ CertData::getVersion( )
    return 0;
  //- CertData::getVersion( )
  }
  
  /**
  */
  public void putCertDisplay(CertDisplay certDisplay)
    throws IllegalStateException,NoSuchMethodException  {
  
  //+ CertData::putCertDisplay(CertDisplay)
    if (isSingleCert())
      ((Composite) getFormat(new Sequence())).add(certDisplay);
    else
      throw new NoSuchMethodException();    
  //- CertData::putCertDisplay(CertDisplay)
  }
  
  /**
  */
  public CertDisplay getCertDisplay()
    throws IllegalStateException,NoSuchMethodException  {
  
  //+ CertData::getCertDisplay( )
    if (isSingleCert())
      return (CertDisplay) ((Composite) getFormat(new Sequence())).
	get(new CertDisplay());
    else
      throw new NoSuchMethodException();
  //- CertData::getCertDisplay( )
  }
  
  /**
  */
  public void putIssuer(PublicKey issuer)
    throws IllegalStateException,NoSuchMethodException  {
  
  //+ CertData::putIssuer(PublicKey)
      if (isSingleCert()) 
	((Composite) getFormat(new Sequence())).add(new Issuer(issuer));
      else
	throw new NoSuchMethodException();
  //- CertData::putIssuer(PublicKey)
  }
  
  /**
  */
  public PublicKey getIssuer()
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::getIssuer( )
    if (isSingleCert())
      return ((Issuer) ((Composite) getFormat(new Sequence())).
	get(new Issuer())).getPublicKey();
    else
      throw new NoSuchMethodException();
  //- CertData::getIssuer( )
  }
  
  /**
  */
  public void putIssuerInfo(URL[] urls) 
       throws NoSuchFieldException,NoSuchMethodException {
  
  //+ CertData::putIssuerInfo(URL)
    if (isSingleCert()) {
	IssuerInfo info = new IssuerInfo();
	for (int i = 0; i < urls.length; i++)
	  info.add(new Uri(urls[i]));
	((Composite) getFormat(new Sequence())).add(info);
    } else
      throw new NoSuchMethodException();
  //- CertData::putIssuerInfo(URL)
  }
  
  /**
  */
  public URL[] getIssuerInfo()
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::getIssuerInfo( )
    if (isSingleCert()) {
      IssuerInfo info = ((IssuerInfo) ((Composite) getFormat(new Sequence())).
	get(new IssuerInfo()));
      URL[] urls = new URL[info.size()];
      for (int i = 0; i < info.size(); i++) 
	urls[i] = ((Uri) info.get(i)).getUrl();
      return urls;
    } else
      throw new NoSuchMethodException();
  //- CertData::getIssuerInfo( )
  }
  
  /**
  */
  public void putSubject(Object subject)
    throws IllegalStateException,NoSuchMethodException  {
  
  //+ CertData::putSubject(Object)
    if (isSingleCert()) {
      if (subject instanceof Subject)
	((Composite) getFormat(new Sequence())).add((Subject) subject);
      else
	((Composite) getFormat(new Sequence())).add(new ObjHash(subject));
    } else
      throw new NoSuchMethodException();
  //- CertData::putSubject(Object)
  }
  
  /**
  */
  public Object getSubject()
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::getSubject( )
    if (isSingleCert())
      return ((Object) ((Composite) getFormat(new Sequence())).
	get(new Subject()));
    else
      throw new NoSuchMethodException();
  //- CertData::getSubject( )
  }
  
  /**
  */
  public void putSubjectInfo(URL[] urls)
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::putSubjectInfo(Url)
    if (isSingleCert()) {
      SubjectInfo info = new SubjectInfo();
      for (int i = 0; i < urls.length; i++)
	info.add(new Uri(urls[i]));
      ((Composite) getFormat(new Sequence())).add(info);
    } else
      throw new NoSuchMethodException();
  //- CertData::putSubjectInfo(Url)
  }
  
  /**
  */
  public URL[] getSubjectInfo()
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::getSubjectInfo( )
    if (isSingleCert()) {
      SubjectInfo info = ((SubjectInfo) ((Composite) getFormat(new Sequence())).
	get(new SubjectInfo()));
      URL[] urls = new URL[info.size()];
      for (int i = 0; i < info.size(); i++) 
	urls[i] = ((Uri) info.get(i)).getUrl();
      return urls;
    } else
      throw new NoSuchMethodException();
  //- CertData::getSubjectInfo( )
  }
  
  /**
  */
  public void putDeleg(Deleg deleg)
    throws IllegalStateException,NoSuchMethodException  {
  
  //+ CertData::putDeleg(Deleg)
      if (isSingleCert())
	((Composite) getFormat(new Sequence())).add(deleg);
      else
	throw new NoSuchMethodException();
  //- CertData::putDeleg(Deleg)
  }
  
  /**
  */
  public Deleg getDeleg()
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::getDeleg( )
      if (isSingleCert())
	return (Deleg) ((Composite) getFormat(new Sequence())).
	  get(new Deleg());
      else
	throw new NoSuchMethodException();
  //- CertData::getDeleg( )
  }
  
  /**
  */
  public void putValidity(Validity valid)
    throws IllegalStateException,NoSuchMethodException  {
  
  //+ CertData::putValidity(Validity)
      if (isSingleCert())
	((Composite) getFormat(new Sequence())).add(valid);
      else
	throw new NoSuchMethodException();
  //- CertData::putValidity(Validity)
  }
  
  /**
  */
  public Validity getValidity()
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::getValidity( )
      if (isSingleCert())
	return (Valid) ((Composite) getFormat(new Sequence())).
	  get(new Valid());
      else
	throw new NoSuchMethodException();
  //- CertData::getValidity( )
  }
  
  /**
  */
  public void putTag(Tag tag)
    throws IllegalStateException,NoSuchMethodException  {
  
  //+ CertData::putTag(Tag)
      if (isSingleCert())
	((Composite) getFormat(new Sequence())).add(tag);
      else
	throw new NoSuchMethodException();
  //- CertData::putTag(Tag)
  }
  
  /**
  */
  public Tag getTag()
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::getTag( )
      if (isSingleCert())
	return (Tag) ((Composite) getFormat(new Sequence())).
	  get(new Tag());
      else
	throw new NoSuchMethodException();
  //- CertData::getTag( )
  }
  
  /**
  */
  public void putComment(Comment comment)
    throws IllegalStateException,NoSuchMethodException  {
  
  //+ CertData::putComment(Comment)
      if (isSingleCert())
	((Composite) getFormat(new Sequence())).add(comment);
      else
	throw new NoSuchMethodException();
  //- CertData::putComment(Comment)
  }
  
  /**
  */
  public Comment getComment()
    throws NoSuchFieldException,NoSuchMethodException  {
  
  //+ CertData::getComment( )
      if (isSingleCert())
	return (Comment) ((Composite) getFormat(new Sequence())).
	  get(new Comment());
      else
	throw new NoSuchMethodException();
  //- CertData::getComment( )
  }
  
} // CertData

// +*************** UNUSED CODE ***************
//+ CertData::putValidity(Validity)
//- CertData::putValidity(Validity)
//+ CertData::getValidity( )
//- CertData::getValidity( )
//+ CertData::putIssuer(PublicKey)
//- CertData::putIssuer(PublicKey)
//+ CertData::putSubject(Object)
//- CertData::putSubject(Object)
//+ CertData::putTag(Tag)
//- CertData::putTag(Tag)
//+ CertData::putDeleg(Boolean)
//- CertData::putDeleg(Boolean)
//+ CertData::putValidity(Validity)
//- CertData::putValidity(Validity)
//+ CertData::putIssuer(PublicKey)
//- CertData::putIssuer(PublicKey)
//+ CertData::putSubject(Object)
//- CertData::putSubject(Object)
//+ CertData::getValidity( )
//- CertData::getValidity( )
//+ CertData::putTag(Tag)
//- CertData::putTag(Tag)
//+ CertData::putDeleg(Boolean)
//- CertData::putDeleg(Boolean)
// -*************** UNUSED CODE ***************
